open Ast 
open Sast
open Types

type symbol_table = {
  parent : symbol_table option;
  variables : Sast.variable_decl list;
  functions: Sast.function_decl list;
}

type trans_env = {
  scope : symbol_table;
}

let rec find_variable (scope : symbol_table) name =
  try
    List.find (fun v -> v.v_name = name) scope.variables
  with Not_found ->
    match scope.parent with
    Some(parent) -> find_variable parent name
    | _ -> raise (Failure("variable  not defined"))

let var_exists scope name =
    List.exists (fun v -> v.v_name = name) scope.variables

let rec find_function (scope : symbol_table) name =
  try
    List.find (fun f -> f.ffname = name) scope.functions
  with Not_found ->
    match scope.parent with
    Some(parent) -> find_function parent name
    | _ -> raise (Failure("function  not defined"))

let func_exists scope name =
    List.exists (fun f -> f.ffname = name) scope.functions

let assign_allowed lt rt = 
  lt = rt

let can_assign (lt:Types.t) rval = 
  let (_, rt) = rval in
  if assign_allowed lt rt  then
    rval
  else
    raise (Failure("Assign Type Error" ))
		
let rec check_int_array b = match b with
  [] -> true
	| h::t -> if snd h = Types.Int then check_int_array t else false
	
	

let can_op lval op rval = 
  let (_, lt) = lval
  and (_, rt) = rval in
  let type_match = (lt = rt) in
  let int_int = (lt= Types.Int && rt=Types.Int) in 
	let int_array =  (lt= Types.Int && rt=Types.Arr) in
	let str_array =  (lt= Types.Str && rt=Types.Arr) in

	let array_int =  (lt= Types.Arr && rt=Types.Int) in
	let array_str =  (lt= Types.Arr && rt=Types.Str) in
	let array_array = (lt= Types.Arr && rt=Types.Arr) in
	
  let array_int =  (lt= Types.Arr && rt=Types.Int) in	
  let result = match op with
      Ast.Add     -> (type_match||array_int||int_array), (if (int_array) then rt else lt)
    | Ast.Sub     -> (int_int), lt
    | Ast.Mult    -> if (int_int||array_array) then (true, lt) else (int_array|| str_array), rt 
    | Ast.Div     -> (int_int), lt
    | Ast.Equal   -> (true, Types.Int)
    | Ast.Neq     -> (true, Types.Int)
    | Ast.Less    -> (int_int), Types.Int
    | Ast.Leq     -> (int_int), Types.Int
    | Ast.Greater -> (int_int), Types.Int
    | Ast.Geq     -> (int_int), Types.Int
    | Ast.And 		-> (int_int), Types.Int
    | Ast.Or  		-> (int_int), Types.Int
		| Ast.Dadd    -> ( array_int), lt
		| Ast.Dsub    -> ( array_int), lt
		| Ast.Dmult   -> ( array_int), lt
		| Ast.Ddiv    -> ( array_int), lt
  in if fst result then
    snd result
  else
    raise (Failure("operand type miss match"))

let translate (globals, funcs) =
  let rec trans_expr env = function
      Ast.Integer(i) -> Sast.Integer(i), Types.Int
    | Ast.String(s)  -> Sast.String(s), Types.Str
    | Ast.Id(n) -> 
        let vdecl = (find_variable env.scope n) in
        Sast.Id(vdecl), vdecl.v_type
    | Ast.Binop(e1, op, e2) ->
        let e1 = trans_expr env e1
        and e2 = trans_expr env e2 in
				let rtype = can_op e1 op e2 in
        Sast.Binop(e1, op, e2), rtype 
    | Ast.Call(n, a) -> 
        let fdecl =  (find_function env.scope n) in
        let types =
          List.map (fun v -> v.v_type) (List.rev fdecl.fformals) in
        let args = 
          List.map (fun s -> (trans_expr env s)) a in
					
        let checked_args = try
            List.map2 can_assign types args
          with Invalid_argument(x) ->
            raise (Failure("invalid number of arguments")) in
        Sast.Call(fdecl, checked_args), fdecl.freturntype 
    | Ast.Assign(n, e) ->
        let vdecl = (find_variable env.scope n) in
        let aval = (trans_expr env e) in
        Sast.Assign(vdecl, (can_assign vdecl.v_type aval)), vdecl.v_type
    | Ast.Noexpr ->
        Sast.Noexpr, Types.Int (*Do we really need this expression?*)
    | Ast.Array(a) ->  let b = List.map (fun v -> trans_expr env v) a in
													if check_int_array b then
														Sast.Array(b), Types.Arr
													else
														 raise (Failure("Int array elements must all be int"))
														
														
													 (*Sast.Array( list.map (fun v -> trans_expr env v) a ), Types.Arr*)
													
	
		| Ast.Index(a,b) -> let a = (find_variable env.scope a)
												and b = trans_expr env b in
													if snd b <> Types.Int then 
														raise (Failure("index must be an integer!")) 
												  else
														 	if a.v_type = Types.Str then
																Sast.Index(a,b),Types.Str
														 	else
															  if a.v_type = Types.Arr then
																	Sast.Index(a,b), Types.Int
																else 
																	raise(Failure("can not perform indexing on this type of variable"))
(*																	
let a = function 
	Sast.Literal i -> match i with
		Integer n -> n
		| _ -> raise (Failure("aaaaa"))
| Sast.Id x -> let v = find_variable env x.v_name in
*)
			    
	in let rec trans_stmt env = function
       Ast.Block(s) -> (*TODO: Allow declaration in new block? *)
        let scope' = {parent = Some(env.scope); variables = []; functions = []}
        in let env' = {env with scope = scope'}
        in let s' = List.map (fun s -> trans_stmt env' s) s
        in Sast.Block(s')
    | Ast.Expr(e) ->
        Sast.Expr(trans_expr env e)
    | Ast.Return(e) ->
        Sast.Return(trans_expr env e)
    | Ast.If (e, s1, s2) ->
        let e' = trans_expr env e
        in Sast.If(can_assign Types.Int e', trans_stmt env s1, trans_stmt env s2)
    | Ast.While (e, s) ->
        let e' = trans_expr env e
        in Sast.While(can_assign Types.Int e', trans_stmt env s)
  in let add_local env v = 
    let evalue = match (var_exists env.scope v.vname) with
        true -> raise (Failure("redeclaration of variable"))
      | false -> 0
    in let new_v = {
            v_name = v.vname;
            v_type = v.vtype;
						v_size = v.vsize;
      }
    in let vars = new_v :: env.scope.variables
    in let scope' = {env.scope with variables = vars}
    in {env with scope = scope'}
  
	in let add_func env f =
    let new_f = match ((var_exists env.scope f.fname) || (func_exists env.scope f.fname)) with
      true -> raise (Failure("redeclaration of function"))
    | false -> {
        freturntype = f.returntype;
        ffname = f.fname;
        fformals = [];
        flocals = [];
        fbody = [];
        parsed = false;
      }
    in let funcs = new_f :: env.scope.functions
    in let scope' = {env.scope with functions = funcs}
    in {env with scope = scope'}
		
  in let trans_func env (f:Ast.func_decl)  =  
    let sf = find_function env.scope f.fname
    in let functions' = List.filter (fun f -> f.ffname != sf.ffname) env.scope.functions
    in let scope' = {parent = Some(env.scope); variables = []; functions = []}
    in let env' = {env with scope = scope'}
    in let env' = List.fold_left add_local env' (f.formals)
    in let formals' = env'.scope.variables
    in let env' = List.fold_left add_local env' (f.locals)
    in let remove v =
      not (List.exists (fun fv -> fv.v_name = v.v_name) formals')
    in let locals' = List.filter remove env'.scope.variables 
    in let body' = List.map (fun f -> trans_stmt env' f) (f.body)
    in let new_f = {
      sf with 
      fformals = formals';
      flocals = locals';
      fbody = body';
      parsed = true;
    }
    in let funcs = new_f :: functions'
    in let scope' = {env.scope with functions = funcs}
    in {env with scope = scope'}
    in let validate_func f =
    let is_return = function
        Sast.Return(e) -> true
      | _ -> false
    in let valid_return = function
        Sast.Return(e) -> if assign_allowed f.freturntype (snd e) then
                            true
                          else
                             raise (Failure(   "Invalid return type " 
                            
                             ))
      | _ -> false
    in let returns = List.filter is_return f.fbody
    in let returns_valid = List.for_all valid_return returns
    in let return_count = List.length returns
    in if (return_count = 0 && f.ffname <> "print" && f.ffname <>"printarray"&& f.ffname <> "printstring" ) then
      raise (Failure( " must return something" ))
    else
      f
   in let make_print t = 
    {
      freturntype = Types.Int;
      ffname = if (t = Types.Str) then "printstring" else "print";
      fformals = [{
        v_name = "val";
        v_type = t;
						v_size=1;
				
      }];
      flocals = [];
      fbody = [];
      parsed = false;
    }

  in let global_scope = {
    parent = None;
    variables = [];
    functions =     [{
      freturntype = Types.Int;
      ffname = "printarray";
      fformals = [{
        v_name = "val";
        v_type = Types.Arr;

				v_size=1;
				
      }];
      flocals = [];
      fbody = [];
      parsed = false;
    }] @ (List.map make_print [Types.Int; Types.Str;]);
  }
 
	 in let genv = {
    scope = global_scope;
  }
  in let genv = List.fold_left add_local genv (List.rev globals)
  in let genv = List.fold_left add_func genv (List.rev funcs)
  in let genv = List.fold_left trans_func genv (List.rev funcs)
  in (genv.scope.variables, List.map validate_func genv.scope.functions)