open Sast
open Bytecode

module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
    function_index : int StringMap.t; (* Index for each function *)
    global_index   : int StringMap.t; (* "Address" for global variables *)
    local_index    : int StringMap.t; (* FP offset for args, locals *)
  }
	let rec remove_print = function 
		 []->[]
		|hd::tl->if (  hd.ffname="print"|| hd.ffname="printarray"||hd.ffname="printstring") then (remove_print tl) else (hd::(remove_print tl))
 
(* val enum : int -> 'a list -> (int * 'a) list *)
let rec enum stride n = function
    [] -> []
  | hd::tl -> 
		if stride > 0 then
		match hd.v_type with
			Types.Int -> (n, hd.v_name) :: enum stride (n + 1) tl
 		|	Types.Str-> (* Here's the question: how many slots need to be allocated to string? *)
						(* To make it simpler, allocate 30 slots for it *)
				(n + hd.v_size-1, hd.v_name) :: enum stride (n + hd.v_size) tl 
    |Types.Arr -> (n + hd.v_size-1, hd.v_name) :: enum stride (n+ hd.v_size) tl
		| 	_ -> raise(Failure ("Undefined type with variable" ^ hd.v_name))
	else
		match hd.v_type with
			Types.Int -> (* Allocate global storage space for an int *)
				(n, hd.v_name) :: enum stride (n + stride) tl
 		|	Types.Str -> (* Here's the question: how many slots need to be allocated to string? *)
						(* To make it simpler, allocate 30 slots for it *)
				(n, hd.v_name) :: enum stride (n + stride * hd.v_size) tl 
 		|	Types.Arr -> 
				(n, hd.v_name) :: enum stride (n+stride * hd.v_size) tl 	
		| 	_ -> raise(Failure ("Undefined type with variable " ^ hd.v_name))

let rec enum_func stride n = function
    [] -> []
  | hd::tl -> (n, hd) :: enum_func stride (n+stride) tl

let get_vari_size a vlist = 
	 List.fold_left (fun a b -> a + (match b.v_type with
										Types.Int -> 1	
									|   Types.Str -> 
b.v_size*1
									|   Types.Arr -> b.v_size*1
									|   _ -> raise(Failure("Error in get_vari_size !!"))
									)) 0 vlist


(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(** Translate a program in AST form into a bytecode program.  Throw an
    exception if something is wrong, e.g., a reference to an unknown
    variable or function *)
let translate (globals, functions) =

  (* Allocate "addresses" for each global variable *)
  let global_indexes = string_map_pairs StringMap.empty (enum 1 0 globals) in

  (* Assign indexes to function names; built-in "print" is special *)
  let built_in_functions = StringMap.add "print" (-1) StringMap.empty in
	 let built_in_functions = StringMap.add "printarray" (-2) built_in_functions in
   let built_in_functions = StringMap.add "printstring" (-3) built_in_functions in
  let function_indexes = string_map_pairs built_in_functions
      (enum_func 1 1 (List.map (fun f -> f.ffname) (remove_print functions))) in

  (* Translate a function in AST form into a list of bytecode statements *)
  let translate env fdecl =
    (* Bookkeeping: FP offsets for locals and arguments *)
    let num_formals = get_vari_size 0 fdecl.fformals
    and num_locals = get_vari_size 0 fdecl.flocals
    and local_offsets = enum 1 1 fdecl.flocals
    and formal_offsets = enum (-1) (-2) fdecl.fformals in
    let env = { env with local_index = string_map_pairs
		  StringMap.empty (local_offsets @ formal_offsets) } in

    let rec expr = function
	Integer i -> [Litin i]
	| String s -> [Litsh  s]
	| Array a -> let rec f l=match l with
	                       []->[]
	                     | hd::tl->expr (fst hd)@f tl in List.rev (f a)
														 
	| Index (x,y) -> (if (x.v_type = Types.Arr || x.v_type = Types.Str) then
								 (try [Litin x.v_size]@[Lfpa (StringMap.find x.v_name env.local_index)]
          with Not_found -> try [Litin x.v_size]@[Loda (StringMap.find x.v_name env.global_index)]
          with Not_found -> raise (Failure ("undeclared variable " ^ x.v_name)))
							else
									raise(Failure ("Indexing performed on wrong type"))			)@expr (fst y)@[GetC]
  | Id s -> if s.v_type = Types.Int then
	  (try [Lfp (StringMap.find s.v_name env.local_index)]
          with Not_found -> try [Lod (StringMap.find s.v_name env.global_index)]
          with Not_found -> raise (Failure ("undeclared variable " ^ s.v_name)))
					else
							if (s.v_type = Types.Arr || s.v_type = Types.Str) then
								 (try [Litin s.v_size]@[Lfpa (StringMap.find s.v_name env.local_index)]
          with Not_found -> try [Litin s.v_size]@[Loda (StringMap.find s.v_name env.global_index)]
          with Not_found -> raise (Failure ("undeclared variable " ^ s.v_name)))
							else
									raise(Failure ("Wrong type of variable"))							
					
  | Binop (e1, op, e2) -> if ((snd e1) = Types.Arr&& op= Ast.Add) then 
				 expr (fst e2)@expr (fst e1)
		            else if ((snd e2) = Types.Arr&& op= Ast.Add) then
															expr (fst e2)@ expr (fst e1)
				else if ((snd e1) = Types.Arr&& op=Ast.Mult&&(snd e2) = Types.Arr) then
				expr (fst e2)@expr (fst e1)
				else if (op=Ast.Dmult||op=Ast.Dadd||op=Ast.Dsub||op=Ast.Ddiv) then
					expr (fst e1) @ [Litin (match (fst e1) with
																				Id s ->s.v_size
																				|Array a -> List.length a)]@expr (fst e2)@[Bin op]
					else
		 expr (fst e1) @ expr (fst e2) @ [Bin op]


  | Assign (s, e) -> expr (fst e) @ (if (s.v_type = Types.Arr || s.v_type = Types.Str) then 
											(try [Litin s.v_size]@[Sfpa (StringMap.find s.v_name env.local_index)]
  	  			with Not_found -> try [Litin s.v_size]@[Stra (StringMap.find s.v_name env.global_index)]
	 					 with Not_found -> raise (Failure ("undeclared variable " ^ s.v_name)))
							else
								if s.v_type = Types.Int then
	  (try [Sfp (StringMap.find s.v_name env.local_index)]
  	  with Not_found -> try [Str (StringMap.find s.v_name env.global_index)]
	  with Not_found -> raise (Failure ("undeclared variable " ^ s.v_name)))
								else
										raise(Failure("can not assign such type"))
		)
  | Call (funcdel, actuals) -> if (funcdel.ffname = "printarray") then
																			expr (fst (List.hd actuals)) @[Litin (match (fst (List.hd actuals)) with 
																																		Array a -> List.length a
|Id s -> s.v_size)
																																		 ]@[Jsr (StringMap.find funcdel.ffname env.function_index)]
																else if (funcdel.ffname = "printstring" && (snd (List.hd actuals)) = Types.Str) then
																			expr (fst (List.hd actuals))@[Litin (match (fst (List.hd actuals)) with 
																																		String s -> String.length s
|Id x -> x.v_size)]@[Jsr (-3)]
															else		
		(try
	  (List.concat (List.map expr (List.map fst (List.rev actuals)))) @
	  [Jsr (StringMap.find funcdel.ffname env.function_index) ]   
        with Not_found -> raise (Failure ("undefined function " ^ funcdel.ffname)))
      | Noexpr -> []

    in let rec stmt = function
	Block sl     ->  List.concat (List.map stmt sl)
      | Expr e       -> expr (fst e) (*@ [Drp]*)
      | Return e     -> expr (fst e) @ [Rts num_formals]
      | If (p, t, f) -> let t' = stmt t and f' = stmt f in
	expr (fst p) @ [Beq(2 + List.length t')] @
	t' @ [Bra(1 + List.length f')] @ f'
      | While (e, b) ->
	  let b' = stmt b and e' = expr (fst e) in
	  [Bra (1+ List.length b')] @ b' @ e' @
	  [Bne (-(List.length b' + List.length e'))]

    in [Ent num_locals] @      (* Entry: allocate space for locals *)
    stmt (Block fdecl.fbody) @  (* Body *)
    [Litin 0; Rts num_formals]   (* Default = return 0 *)

  in let env = { function_index = function_indexes;
		 global_index = global_indexes;
		 local_index = StringMap.empty } in

  (* Code executed to start the program: Jsr main; halt *)
  let entry_function = try
    [Jsr (StringMap.find "main" function_indexes); Hlt]
  with Not_found -> raise (Failure ("no \"main\" function"))
  in
    
  (* Compile the functions *)
  let func_bodies = entry_function :: List.map (translate env) functions in

  (* Calculate function entry points by adding their lengths *)
  let (fun_offset_list, _) = List.fold_left
      (fun (l,i) f -> (i :: l, (i + List.length f))) ([],0) func_bodies in
  let func_offset = Array.of_list (List.rev fun_offset_list) in

  { size_globals = get_vari_size 0 globals;
    (* Concatenate the compiled functions and replace the function
       indexes in Jsr statements with PC values *)
    text = Array.of_list (List.map (function
	Jsr i when i > 0 -> Jsr func_offset.(i)
      | _ as s -> s) (List.concat func_bodies))
  }